from typing import Optional
from decouple import Config

from nornir import InitNornir
from nornir.core import Nornir

from ..credentials import update_nornir_credentials
from .logging import configure_nornir_logging, DEFAULT_LOGFILE
from .connection_options import configure_connection_options

CHUNK_SIZE = 5
NUM_WORKERS = 12


NETBOX_DEVICE_ROLES = [
    "av",
    "access-layer-switch",
    "bin",
    "core",
    "data-center",
    "distribution",
    "out-of-band",
    "security",
    "umd",
]

NETBOX_ILAB_DEVICE_ROLES = [
    "access",
    "agg",
    "bgw",
    "core",
    "data-center",
    "distribution",
    "legacy-bin",
    "legacy-core",
    "legacy-data-center",
    "legacy-distribution",
    "ngfw",
    "pe",
]


def nornir_setup(
    cfg:Config,
    log_level: str = "DEBUG",
    log_globally: Optional[bool] = False,
    log_to_console: Optional[bool] = False,
    log_file: str = DEFAULT_LOGFILE,
    device_filter: Optional[str] = None,
    role_filter: Optional[str] = None,
) -> Nornir:
    """
    Initializes Nornir to point at netbox, and to only care about active
    devices tied to a specific subset of device roles.
    Sets up logging. Populates default and custom passwords from cyberark. Returns
    customized Nornir instance
    """

    configure_nornir_logging(log_level, log_globally, log_file, log_to_console)

    nb_url = cfg.get("NB_URL")

    filter_params = {"status": "active", "has_primary_ip": "True"}

    # Restrict what the netbox inventory plugin pulls if it was indicated
    # on the CLI
    if device_filter:
        filter_params["name"] = device_filter
    elif role_filter:
        filter_params["role"] = [role_filter]
    elif "ilab" in nb_url:
        filter_params["role"] = NETBOX_ILAB_DEVICE_ROLES
    else:
        filter_params["role"] = NETBOX_DEVICE_ROLES

    # Nornir initialization
    nr = InitNornir(
        runner={
            "plugin": "multiprocess",
            "options": {
                "num_workers": NUM_WORKERS,
            },
        },
        inventory={
            "plugin": "NetBoxInventory2",
            "options": {
                "nb_url": nb_url,
                "nb_token": cfg.get("NB_TOKEN"),
                "filter_parameters": filter_params,
                "ssl_verify": False,
            },
        },
        logging={
            "enabled": False,
        },
    )

    update_nornir_credentials(nr, cfg.get("CYBERARK_ENV_FILE"))
    configure_connection_options(nr)

    return nr



"""
Note - the Cyberark class below is adapted from one that Jeff Hagley wrote
and placed in umnet-scripts. Instead of augmenting that class, we've created
a new one here so that this package isn't dependent on the umnet-scripts
package, which has many dependencies of its own.
"""

import sys
from typing import Optional
from dataclasses import dataclass
from nornir.core import Nornir
from nornir.core.inventory import ConnectionOptions

from decouple import Config, RepositoryEnv, config
import requests
import urllib3

urllib3.disable_warnings()

CYBERARK_TIMEOUT=30

class CyberarkLookupError(Exception):
    pass

class CyberarkTimeoutError(Exception):
    pass


class CyberarkAuthError(Exception):
    pass


@dataclass
class CyberarkEnvironment:
    name: str
    cert: tuple[str, str]
    appid: str
    safe: str


class Cyberark:
    """
    Class that interacts with the cyberark API.
    """

    def __init__(
        self, env_file: str = "cyberark.env", environment: Optional[str] = "UMNET"
    ):

        self.env_config = Config(RepositoryEnv(env_file))
        self.url_base = self.env_config.get("URL_BASE")
        self.url_aim_endpoint = self.env_config.get("URL_AIM_ENDPOINT")
        self.headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        self.set_environment(environment)

    def set_environment(self, environment: str):
        environment = environment.upper()

        if environment == "UMNET":
            self.env = CyberarkEnvironment(
                environment,
                (self.env_config.get("CERT"), self.env_config.get("KEY")),
                self.env_config.get("APPID"),
                self.env_config.get("SAFE"),
            )
        elif environment == "NSO":
            self.env = CyberarkEnvironment(
                environment,
                (self.env_config.get("NSO_CERT"), self.env_config.get("NSO_KEY")),
                self.env_config.get("NSO_APPID"),
                self.env_config.get("NSO_SAFE"),
            )
        elif environment == "DEARBORN":
            self.env = CyberarkEnvironment(
                environment,
                (
                    self.env_config.get("DEARBORN_CERT"),
                    self.env_config.get("DEARBORN_KEY"),
                ),
                self.env_config.get("DEARBORN_APPID"),
                self.env_config.get("DEARBORN_SAFE"),
            )

        else:
            raise ValueError(f"Unknown cyberark environment {environment}")

    def query_cyberark(
        self, password_entity: str, environment: Optional[str] = None
    ) -> str:
        """
        Queries cyberark for a specific entity
        """
        if sys.version.startswith("3.8"):
            requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = (  # pylint: disable=no-member
                "ALL:@SECLEVEL=1"
            )
        if environment:
            self.set_environment(environment)

        aim_url = (
            self.url_base
            + self.url_aim_endpoint
            + self.env.appid
            + self.env.safe
            + "&Object="
            + password_entity
        )
        try:
            acct = requests.get(
                aim_url, headers=self.headers, cert=self.env.cert, verify=True, timeout=CYBERARK_TIMEOUT
            )
        except requests.exceptions.ReadTimeout:
            raise CyberarkTimeoutError(f"Cyberark query for {password_entity} timed out after {CYBERARK_TIMEOUT} seconds")

        if acct.status_code == 404:
            raise CyberarkLookupError(
                f"No password found for {password_entity} environment {self.env.name}"
            )
        if acct.status_code == 403:
            raise CyberarkAuthError(f"Could not auth to cyberark for {self.env.name}")

        result = acct.json()

        if acct.status_code != 200 or not result or "Content" not in result:
            raise CyberarkLookupError(
                f"Invalid response from cyberark for {password_entity} environment {self.env.name}"
            )

        return result["Content"]

    def lookup_enable(self, host: str = "", environment: Optional[str] = None) -> str:
        """
        Looks up an 'enable' password for a specific host
        """
        if environment:
            self.set_environment(environment)

        return self.query_cyberark(f"{host}_enable")

    def lookup_username_password(
        self,
        username: str,
        host: Optional[str] = None,
        environment: Optional[str] = None,
    ) -> str:
        """
        Looks up the password for a specific username, optionally on a specific host.
        If there's no password found for that specific host, looks for that
        username for *all* hosts
        """
        if environment:
            self.set_environment(environment)

        # host passwords without usernames are stored in cyberark as 'name_user_name'
        if host is None:
            entity = f"{username}_user_{username}"
        else:
            entity = f"{host}_{username}"

        try:
            return self.query_cyberark(entity)
        except CyberarkLookupError:
            return self.query_cyberark(f"_{username}")



import logging
import traceback
from nornir.core.task import Result, Task
from nornir.core.exceptions import NornirSubTaskError
from sqlalchemy.engine import Engine

from umnet_napalm import get_network_driver

from ..utils import get_device_cmd_list
from ..mappers.save_to_file import SaveResult
from ..mappers.save_to_db import ResultsToDb

from .nornir_setup import NUM_WORKERS
from sqlalchemy import create_engine

logger = logging.getLogger(__name__)


def process_device(
    task: Task,
    cmd_list: list,
    cmd_map: dict,
    db_url:str,
) -> Result:
    """
    Run a set of commands against a specific device based on the
    command mapper and an optional restricted list of commands.
    If no list is provided all commands in the mapper will be run
    """
    logger.debug(f"In process device task for {task.host}")
    failed = False
    db_engine = create_engine(
        db_url, pool_size=NUM_WORKERS, max_overflow=NUM_WORKERS + 10
    )
    # first figure out which commands to run against this host
    cmd_list = get_device_cmd_list(cmd_map, task.host, cmd_list_filter=cmd_list)

    logger.info(f"Running getters on {task.host.name}: {cmd_list}")
    if not cmd_list:
        logger.info(f"No commands to execute for {task.host.name}")
        return Result(host=task.host, result="No commands to execute")

    # finding that for multiprocessing it's cleaner to use the context
    # manager - by this we mean 'with network_driver as device' - because
    # connection failures that raise exceptions aren't well tolerated.
    network_driver = get_network_driver(task.host.platform)
    parameters = {
        "hostname": task.host.hostname,
        "username": task.host.username,
        "password": task.host.password,
        "optional_args": {},
    }
    parameters.update(task.host.connection_options["umnet_napalm"].extras)

    getter_results = {}
    try:
        with network_driver(**parameters) as device:

            for cmd, options in cmd_list.items():
                method = getattr(device, cmd_map[cmd]["getter"])
                getter_results[cmd] = method(**options)

    except Exception as e:
        tb = traceback.format_exc()
        logger.error(
            "Host %r getters failed with traceback:\n%s",
            task.host.name,
            tb,
        )
        return Result(host=task.host, result=tb, error=str(e), failed=True)

    # for each command save data using the appropriate methods based
    # on the command map
    for cmd, result in getter_results.items():

        if "save_to_db" in cmd_map[cmd]:
            try:
                task.run(
                    name=f"{cmd}_save_to_db",
                    task=update_table,
                    result=result,
                    engine=db_engine,
                    mapper=cmd_map[cmd]["save_to_db"],
                )
            except NornirSubTaskError:
                failed = True

        if "save_to_file" in cmd_map[cmd]:
            try:
                task.run(
                    name=f"{cmd}_save_to_file",
                    task=save_to_file,
                    result=result,
                    repo_path=cmd_map[cmd]["save_to_file"]["destination"],
                    mapper=cmd_map[cmd]["save_to_file"]["mapper"],
                )
            except NornirSubTaskError:
                failed = True

    return Result(host=task.host, result="Completed", failed=failed)


def update_table(
    task: Task,
    result: Result,
    engine: Engine,
    mapper: ResultsToDb,
) -> Result:
    """
    Generic task for taking the results from a previous "napalm_get"
    task and saving it to the database
    """
    mapper(task.host.name, task.host.hostname, result, engine)

    return Result(host=task.host)


def save_to_file(
    task: Task,
    result: dict,
    repo_path: str,
    mapper: SaveResult,
) -> Result:
    """
    Task for saving to a file
    """
    mapper(task.host, repo_path, result)

    return Result(host=task.host)



import sys
import re
import argparse
from datetime import datetime
import socket
import logging


from .nornir import nornir_setup

from .nornir.tasks import process_device
from .nornir.processors import (
    AgadorProcessor,
    TraceFile,
)
from .utils import git_update, parse_command_map, get_config_settings


logger = logging.getLogger(__name__)

def main():

    parser = argparse.ArgumentParser(description="Run agador")

    parser.add_argument("--cfg-file", help="Configuration file. You can also set this as AGADOR_CONFIG in your environennt")

    filter_args = parser.add_mutually_exclusive_group()
    filter_args.add_argument(
        "--device", help="Restrict the update to a particular device"
    )
    filter_args.add_argument(
        "--role", help="Restrict the update to a specific netbox device role"
    )

    parser.add_argument(
        "--cmds",
        nargs="*",
        help="Limit update to a subset of tasks/commands",
    )


    log_args = parser.add_mutually_exclusive_group()
    log_args.add_argument("-l", "--log-level", help="Set log level for agador only")
    log_args.add_argument("-L", "--LOG-LEVEL", help="set log level for all libraries")
    parser.add_argument("--echo", action="store_true", help="echo logfile to stdout")
    parser.add_argument("--trace", action="store_true", help="Save device session logs")

    args = parser.parse_args()

    # get config settings
    cfg = get_config_settings(args.cfg_file)
    cmd_map = parse_command_map(cfg)

    # input validation
    if args.cmds:
        for cmd in args.cmds:
            if cmd not in cmd_map:
                print(f"ERROR: {cmd} not in command map! valid commands: {','.join(cmd_map.keys())}")
                sys.exit(1)
            
    if args.trace and not args.device:
        sure = input(
            "Trace output is for debugging only. Are you sure you want to save session logs for ALL devices (y/n)? "
        )
        if re.match(r"[Yy]", sure):
            print(f"Fine. Logs will be saved at {cfg.get('LOG_DIR')}")
        else:
            print("Good choice! Turning trace off.")
            args.trace = False

    # log level
    if args.log_level:
        log_level = args.log_level
    elif args.LOG_LEVEL:
        log_level = args.LOG_LEVEL
    else:
        log_level = logging.INFO


    # initialize nornir
    logger.info("initializing nornir")
    nr = nornir_setup(
        cfg=cfg,
        log_level=log_level,
        log_globally=bool(args.LOG_LEVEL),
        log_to_console=args.echo,
        device_filter=args.device,
        role_filter=args.role,
    )

    logger.info(
        "Nornir initialization complete - inventory has %s items",
        len(nr.inventory.hosts.keys()),
    )

    if not (nr.inventory.hosts.keys()):
        logger.error("No matching hosts found in netbox inventory!")
        sys.exit(1)

    # run commands against all our devices
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    results_logfile = f"{cfg.get('LOG_DIR')}/results_{timestamp}"

    # setting up processors
    processors = [
        AgadorProcessor(len(nr.inventory.hosts), cfg.get("EMAIL_ADDRESS"), cfg.get("EMAIL_RECIPIENTS"), results_logfile, args.echo)
    ]
    if args.trace:
        processors.append(TraceFile(cfg.get("LOG_DIR")))

    logger.debug("Starting run...")
    nr.with_processors(processors).run(
        task=process_device,
        cmd_list=args.cmds,
        cmd_map=cmd_map,
        db_url=cfg.get("DB_URL"),
        on_failed=True,
    )

    # add/commit/push git changes if applicable
    logger.debug("git updates")
    formatted_date = datetime.strftime(datetime.now(), "%a %d %b %Y, %I:%M%p")
    commit_msg = f"Agador backup from {socket.gethostname()} at {formatted_date}"

    for cmd, data in cmd_map.items():
        if not data.get("save_to_file"):
            continue
        if args.cmds and cmd not in args.cmds:
            continue
        git_update(data["save_to_file"]["destination"], commit_msg)

if __name__ == "__main__":
    main()


# pylint: disable=too-few-public-methods, not-callable
from typing import Dict, Any
from datetime import datetime, timedelta
import logging

from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy import select, Column
from sqlalchemy.sql import sqltypes

from nornir.core.task import Result

from umnet_db.models import *

from umnet_napalm.utils import abbr_interface

logger = logging.getLogger(__name__)

# Maping table columns to custom conversion
CUSTOM_COLUMN_CASTS = {
    # age and uptime from umnet_napalm come in as integers
    "age": lambda x: timedelta(seconds=x),
    "uptime": lambda x: timedelta(seconds=x),
}


class NapalmToDBError(Exception):
    pass


class ResultsToDb:
    """
    Generic class that maps nornir results to a particular table
    in the database
    """

    # db table to map to
    db: Base

    # name of column that holds the device name. For the Device table
    # this is 'name', for everything else it's device
    _device_col: str = "device"

    # when we update the table are we merging the device data or overwriting?
    # we almost always want to overwrite.
    _overwrite = True

    def __init__(self, hostname: str, host_ip: str, results: Result, engine: Engine):
        self._hostname = hostname
        self._host_ip = host_ip
        self._results = results
        self._e = engine
        self._timestamp = datetime.now()

        self.update()

    @property
    def pks(self) -> tuple:
        """
        List of column names that are primary keys for this table, it
        """
        return tuple(c.name for c in self.db.__table__.primary_key.columns)

    def _pk(self, row: Base) -> tuple:
        """
        extracts primary key data from a row object
        """
        return tuple(getattr(row, pk) for pk in self.pks)

    def update(self) -> bool:
        """
        Open a session with the database. Creates new row objects
        then either replace existing device's rows in
        the table with new ones or merge them in.

        Returs a bool that tells us if we modified the db or not
        Right now we're assuming any failure will result in an exception -
        if we get to the "commit" at the end that is a success.
        """

        with Session(self._e) as session:

            # pulling current and new rows
            logger.debug(
                "Getting current entries for %s %s",
                self._hostname,
                self.db.__tablename__,
            )
            curr_rows = self._get_current_rows(session)
            # logger.debug([r for r in curr_rows.values()])
            logger.debug("Retrieved %s current entries", len(curr_rows))

            logger.debug("Generating rows from given data")
            new_rows = self._create_rows()
            # logger.debug([r for r in new_rows.values()])
            logger.debug("Generated %s rows", len(new_rows))

            # now we must compare the keys of the new rows to the current ones
            for pk, new_data in new_rows.items():

                # adding rows that don't exist
                if pk not in curr_rows:
                    logger.debug(f"Adding new pk {pk}")
                    session.add(new_data)

                # for existing rows, copy over all columns except for 'first seen'
                else:
                    logger.debug(f"updating existing row {curr_rows[pk]}")
                    for k, v in new_data.as_dict().items():
                        if k != "first_seen":
                            setattr(curr_rows[pk], k, v)

            # if we are overwriting, we want to blow away any existing row
            # not in the current data
            if self._overwrite:
                for pk, curr_data in curr_rows.items():
                    if pk not in new_rows:
                        logger.debug(f"Deleting {pk} {curr_data}")
                        session.delete(curr_data)

            logger.debug("Committing data")
            session.commit()

            # the boolean returned indicates that we modified the database
            return True

    def _create_rows(self) -> Dict[tuple, Base]:
        """
        Generic "create rows" that works when there's a 1:1 mapping between
        the getter results and the DB.
        """
        rows = {}
        for result in self._results:
            row = self.db()
            self._result_to_row(row, result)
            rows[self._pk(row)] = row

        return rows

    def _result_to_row(self, row: Base, result: dict, mapping: dict = None):
        """
        generic "save items from this napalm results dictionary into the row's column
        attributes. If the results dictionary's keys don't map directly to
        the table attributes, provide a mapping so we know what maps to what.
        """
        if mapping:
            for k, v in mapping.items():
                col = getattr(self.db.__table__.columns, k)
                setattr(row, k, self._cast_col(col, result[v]))
        else:
            for k, v in result.items():
                col = getattr(self.db.__table__.columns, k, None)
                if col is not None:
                    setattr(row, k, self._cast_col(col, v))

        # adding first seen and last updated fields
        row.first_seen = self._timestamp
        row.last_updated = self._timestamp

        # also - set the 'device' attribute in the row if it's not already set
        # we don't get the device name from our napalm results
        if "device" in row.__table__.columns and not row.device:
            row.device = self._hostname

    def _get_current_rows(self, s: Session) -> Dict[tuple, Base]:
        """
        Pulls the current data for a particular device for this table
        """
        rows = s.scalars(
            select(self.db).where(getattr(self.db, self._device_col) == self._hostname)
        )
        if not rows:
            return {}
        return {self._pk(row): row for row in rows}

    def _cast_col(self, col: Column, entry: Any) -> Any:
        """
        Attempts to cast entries from napalm to the right python type.
        """
        # empty strings for non-string, nullable columns should get converted to None type
        if not (isinstance(col.type, sqltypes.String)) and not entry and col.nullable:
            return None

        # converting to python type if it's not that way already
        if entry and not (isinstance(entry, col.type.python_type)):

            # special handling of certain columns
            for cast_col_name, cast_function in CUSTOM_COLUMN_CASTS.items():
                if col.name == cast_col_name:
                    return cast_function(entry)

            # sqlalchemy types have an attribute 'python type'
            # that we can use for casting
            return col.type.python_type(entry)

        return entry


class UpdateDevice(ResultsToDb):
    """
    Takes "get facts" and maps it to the Device table
    """

    db = Device
    _device_col = "name"
    _overwrite = False

    def _create_rows(self) -> Dict[tuple, Device]:
        """
        Maps napalm "get facts" and a the hostname/ip from nornir
        to a row in the device table
        """
        device = self.db()

        device.name = self._hostname
        device.ip = self._host_ip

        get_facts_mapping = {
            "version": "os_version",
            "vendor": "vendor",
            "model": "model",
            "serial": "serial_number",
            "uptime": "uptime",
        }

        self._result_to_row(device, self._results, get_facts_mapping)
        return {self._pk(device): device}


class UpdateNeighbor(ResultsToDb):
    """
    Takes "show lldp neighbors" and maps it to the Device table
    """

    db = Neighbor

    def _create_rows(self) -> Dict[tuple, Neighbor]:

        rows = {}
        for port, neighs in self._results.items():
            for neigh in neighs:
                new_neigh = self.db(
                    device=self._hostname,
                    port=self._cleanup_interface(port),
                    remote_device=self._cleanup_hostname(neigh["hostname"]),
                    remote_port=self._cleanup_interface(neigh["port"]),
                    first_seen=self._timestamp,
                    last_updated=self._timestamp,
                )

                rows[self._pk(new_neigh)] = new_neigh

        return rows

    def _cleanup_hostname(self, hostname: str) -> str:
        return hostname.split(".")[0].lower()

    def _cleanup_interface(self, interface: str) -> str:
        """
        Abbreviates cisco interfaces and converts logical to physical for junos
        """
        # want physical interface not logical
        if "." in interface:
            interface = interface.split(".")[0]

        return abbr_interface(interface)


class UpdateLAG(ResultsToDb):
    """
    Takes "get_lag_interfaces" and maps it to the LAG table.
    Note that the 'member' column of the LAG table is required, and
    as a result we're not going to save data on LAGs with no configured
    members.
    """

    db = Lag

    def _create_rows(self) -> Dict[tuple, Lag]:
        rows = {}
        for lag_name, lag_data in self._results.items():
            for member_name, member_data in lag_data.get("members", {}).items():
                new_member = self.db(
                    device=self._hostname,
                    member=member_name,
                    parent=lag_name,
                    protocol=lag_data["protocol"],
                    admin_up=member_data["admin_up"],
                    oper_up=member_data["oper_up"],
                    vpc_id=lag_data["mlag_id"],
                    peer_link=lag_data["peer_link"],
                    first_seen=self._timestamp,
                    last_updated=self._timestamp,
                )
                rows[self._pk(new_member)] = new_member

        return rows


class UpdateARP(ResultsToDb):
    """
    Takes "get_arp_table" and maps it to the ARP table. Note that
    if there's an existing ARP entry, we should just be updating
    the 'last seen' value
    """

    db = ARP


class UpdateIPInterface(ResultsToDb):
    """
    Takes "get_ip_interfaces" and maps it to the IPInterface table
    """

    db = IPInterface


class UpdateRoute(ResultsToDb):
    """
    Takes "get_active_routes" and maps it to the Routes table
    """

    db = Route


class UpdateMPLS(ResultsToDb):
    """
    Takes "get_mpls_switching" and maps it to the MPLS table.
    Note that we actually need to query the DB for the
    device's routes in the default table to resolve
    next hops first!
    """

    db = MPLS


class UpdateVNI(ResultsToDb):
    """
    Takes "get_vni_information" and maps it to the VNI table
    """

    db = VNI


class UpdateInventory(ResultsToDb):
    """
    Takes "get_inventory" and maps it to the inventory table
    """

    db = Inventory
